from luna.common.utils import (
    generate_uuid,
    get_absolute_path,
    validate_params,
    cli_runner,
    expand_inputs,
)
import pytest
import os
import yaml


def test_generate_uuid():
    uuid = generate_uuid(
        "file:./pyluna-common/tests/luna/common/test_config.yml", ["FEATURE", "label"]
    )

    assert uuid.startswith("FEATURE-label-")


def test_get_absolute_path():
    absolute_path = get_absolute_path(
        __file__, "../data_ingestion_template_invalid.yml"
    )
    assert absolute_path.startswith("/")
    assert absolute_path.endswith(
        "pyluna-common/tests/luna/data_ingestion_template_invalid.yml"
    )


def test_validate_params__simple():
    params_list = [("input", str), ("threshold", float)]

    given_params = {
        "input": "data.json",
        "threshold": 0.5,
        "extra": 1,
    }
    out_params = validate_params(given_params, params_list)

    assert out_params == {
        "input": "data.json",
        "threshold": 0.5,
    }


def test_validate_params__casting():
    params_list = [("input", str), ("threshold", float)]

    given_params = {
        "input": "data.json",
        "threshold": "0.5",  # Cast string
        "extra": 1,
    }
    out_params = validate_params(given_params, params_list)

    assert out_params == {
        "input": "data.json",
        "threshold": 0.5,
    }


def test_validate_params__missing_value():
    with pytest.raises(RuntimeError):

        params_list = [("input", str), ("threshold", float)]

        given_params = {
            "threshold": 0.5,  # Cast string
            "extra": 1,
        }
        out_params = validate_params(given_params, params_list)

        assert out_params == {
            "input": "data.json",
            "threshold": 0.5,
        }


def test_validate_params__wrong_value():
    with pytest.raises(RuntimeError):

        params_list = [("input", str), ("threshold", float)]

        given_params = {
            "input": "data.json",
            "threshold": "Five",  # Cast string
            "extra": 1,
        }
        out_params = validate_params(given_params, params_list)

        assert out_params == {
            "input": "data.json",
            "threshold": 0.5,
        }


def test_validate_params_dict():
    params_list = [("input1", dict), ("input2", dict)]

    given_params = {
        "input1": "{'key':'not_json_format'}",
        "input2": {"key": "not_json_format"},
    }

    out_params = validate_params(given_params, params_list)

    assert out_params == {
        "input1": {"key": "not_json_format"},
        "input2": {"key": "not_json_format"},
    }


def test_expand_inputs_explicit():
    given_params = {
        "input_text_file": "pyluna-common/tests/luna/common/testdata/simple_output_directory/result.txt",
        "threshold": "Five",
        "extra": 1,
    }

    kwargs, keys = expand_inputs(given_params)

    assert (
        kwargs["input_text_file"]
        == "pyluna-common/tests/luna/common/testdata/simple_output_directory/result.txt"
    )


def test_expand_inputs_implicit():

    given_params = {
        "input_text_file": "pyluna-common/tests/luna/common/testdata/simple_output_directory/",
        "threshold": "Five",
        "extra": 1,
    }

    kwargs, keys = expand_inputs(given_params)

    assert (
        kwargs["input_text_file"]
        == "pyluna-common/tests/luna/common/testdata/simple_output_directory/result.txt"
    )


def test_expand_inputs_implicit_but_missing():
    with pytest.raises(RuntimeError):
        given_params = {
            "input_csv_file": "pyluna-common/tests/luna/common/testdata/simple_output_directory/",
            "threshold": "Five",
            "extra": 1,
        }

        expand_inputs(given_params)


def test_cli_runner(tmp_path):
    def simple_transform(input_text_file, message, output_dir, username):
        with open(input_text_file, "r") as fp:
            old_message = fp.read()

        new_message = old_message + "\n" + message
        with open(output_dir + "/message.txt", "w") as fp:
            fp.write(new_message)

        properties = {
            "num_characters": len(new_message),
        }

        return properties

    cli_kwargs = {
        "input_text_file": "pyluna-common/tests/luna/common/testdata/simple_output_directory/result.txt",
        "output_dir": tmp_path,
        "message": "Hello to you too!",
        "username": "my-user",
    }

    _params_ = [
        ("input_text_file", str),
        ("output_dir", str),
        ("message", str),
        ("username", str),
    ]
    cli_runner(cli_kwargs, _params_, simple_transform)

    assert os.path.exists(str(tmp_path) + "/metadata.yml")
    assert os.path.exists(str(tmp_path) + "/message.txt")

    with open(str(tmp_path) + "/metadata.yml") as fp:
        metadata = yaml.safe_load(fp)
        assert metadata["num_characters"] == 24


def test_cli_runner_reruns(tmp_path):
    def simple_transform(input_text_file, message, output_dir):
        with open(input_text_file, "r") as fp:
            old_message = fp.read()

        new_message = old_message + "\n" + message
        with open(output_dir + "/message.txt", "w") as fp:
            fp.write(new_message)

        properties = {
            "num_characters": len(new_message),
        }

        return properties

    cli_kwargs = {
        "input_text_file": "pyluna-common/tests/luna/common/testdata/simple_output_directory/result.txt",
        "output_dir": tmp_path,
        "message": "Hello to you too!",
    }

    _params_ = [("input_text_file", str), ("output_dir", str), ("message", str)]
    cli_runner(cli_kwargs, _params_, simple_transform)

    cli_kwargs_rerun = {
        "method_param_path": str(tmp_path) + "/metadata.yml",
        "output_dir": str(tmp_path) + "rerun",
    }

    cli_runner(cli_kwargs_rerun, _params_, simple_transform)

    with open(str(tmp_path) + "rerun" + "/metadata.yml") as fp:
        metadata = yaml.safe_load(fp)
        assert metadata["num_characters"] == 24



import os
import json
import yaml
import logging

from filehash import FileHash
from importlib import import_module
from io import BytesIO
from typing import Callable, List
from luna.common.CodeTimer import CodeTimer
import itertools

import shutil

import pandas as pd
from pathlib import Path

import requests
from functools import partial 
import urllib

import warnings

logger = logging.getLogger(__name__)

# Distinct types that are actually the same (effectively)
TYPE_ALIASES = {"itk_geometry": "itk_volume"}

# Sensitive cli inputs
MASK_KEYS = ["username", "user", "password", "pw"]

def to_sql_field(s):
    filter1 = s.replace(".", "_").replace(" ", "_")
    filter2 = "".join(e for e in filter1 if e.isalnum() or e == "_")
    return filter2


def to_sql_value(s):
    if isinstance(s, str):
        return f"'{s}'"
    if not s == s:
        return "Null"
    if s is None:
        return "Null"
    else:
        return f"{s}"


def clean_nested_colname(s):
    """
    Removes map name for MapType columns.
    e.g. metadata.SeriesInstanceUID -> SeriesInstanceUID
    """
    return s[s.find(".") + 1 :]


def generate_uuid(path, prefix):
    """
    Returns hash of the file given path, preceded by the prefix.
    :param path: file path e.g. file:/path/to/file
    :param prefix: list e.g. ["SVGEOJSON","default-label"]
    :return: string uuid
    """
    posix_file_path = path.split(":")[-1]

    rec_hash = FileHash("sha256", chunk_size=65536).hash_file(posix_file_path)
    prefix.append(rec_hash)
    return "-".join(prefix)


def rebase_schema_numeric(df):
    """
    Tries to convert all columns in a dataframe to numeric types, if possible, with integer types taking precident

    Note: this is an in-place operation

    Args:
        df (pd.DataFrame): dataframe to convert columns
    """
    for col in df.columns:
        if df[col].dtype != object:
            continue

        df[col] = df[col].astype(float, errors="ignore")
#        df[col] = df[col].astype(int, errors="ignore") # This was converting small floats to integers...


def generate_uuid_binary(content, prefix):
    """
    Returns hash of the binary, preceded by the prefix.
    :param content: binary
    :param prefix: list e.g. ["FEATURE"]
    :return: string uuid
    """
    warnings.warn("generate_uuid_binary() should not be used anymore, the UUIDs generated are not valid!")

    content = BytesIO(content)

    uuid = '00000000'
    prefix.append(uuid)
    return "-".join(prefix)


def generate_uuid_dict(json_str, prefix):
    """
    Returns hash of the json string, preceded by the prefix.
    :param json_str: str representation of json
    :param prefix: list e.g. ["SVGEOJSON","default-label"]
    :return: v
    """
    json_bytes = json.dumps(json_str).encode("utf-8")
    warnings.warn("generate_uuid_dict() should not be used anymore, the UUIDs generated are not valid!")

    uuid = '00000000'
    prefix.append(uuid)
    return "-".join(prefix)


def does_not_contain(token, value):
    """
    Validate that `token` is not a substring of `value`

    :param: token: string e.g. : | .
    :param: value: dictionary, list, or str
    """
    if isinstance(value, str):
        if token in value:
            raise ValueError(f"{value} cannot contain {token}")

    if isinstance(value, list):
        if any([token in v for v in value]):
            raise ValueError(str(value) + f" cannot contain {token}")

    if isinstance(value, dict):
        if any(
            [
                isinstance(key, str)
                and token in key
                or isinstance(val, str)
                and token in val
                for key, val in value.items()
            ]
        ):
            raise ValueError(str(value) + f" cannot contain {token}")

    return True


def replace_token(token, token_replacement, value):
    """
    Replace `token` with `token_replacement` in `value`

    :param: token: string e.g. : | .
    :param: token_replacement: string e.g. _ -
    :param: value: dictionary, list, or str
    """
    if isinstance(value, str):
        return value.replace(token, token_replacement)

    if isinstance(value, list):
        new_value = []
        for v in value:
            new_value.append(v.replace(token, token_replacement))
        return new_value

    if isinstance(value, dict):
        new_value = {}
        for key, val in value.items():
            new_key, new_val = key, val
            if isinstance(key, str):
                new_key = key.replace(token, token_replacement)
            if isinstance(val, str):
                new_val = val.replace(token, token_replacement)
            new_value[new_key] = new_val

        return new_value

    return value


def grouper(iterable, n):
    """Turn an iterable into an iterable of iterables

    'None' should not be a member of the input iterable as it is removed to handle the fillvalues

    Args:
        iterable (iterable): an iterable
        n (int): sie of chunks
        fillvalue

    Returns:
        iterable[iterable]
    """
    args = [iter(iterable)] * n
    return [
        [entry for entry in iterable if entry is not None]
        for iterable in itertools.zip_longest(*args, fillvalue=None)
    ]


def get_method_data(cohort_id, method_id):
    """
    Return method dict

    :param: cohort_id: string
    :param: method_id: string
    """

    method_dir = os.path.join(os.environ["MIND_GPFS_DIR"], "data", cohort_id, "methods")
    with open(os.path.join(method_dir, f"{method_id}.json")) as json_file:
        method_config = json.load(json_file)["params"]
    return method_config


def get_absolute_path(module_path, relative_path):
    """Given the path to a module file and the path, relative to the module file, of another file
    that needs to be referenced in the module, this method returns the absolute path of the file
    that needs to be referenced.

    This method makes it possible to resolve absolute paths to files in any environment a
    module and the referenced files are deployed to.

    :param module_path path to the module. Use '__file__' from the module.
    :param relative_path path to the file that needs to be referenced by the module. The path must
    be relative to the module.
    :return absolute path to file with the specified relative_path
    """
    path = os.path.join(os.path.dirname(module_path), relative_path)

    # resolve any back-paths with ../ to simplify absolute path
    return os.path.realpath(path)


def validate_params(given_params: dict, params_list: List[tuple]):
    """Ensure that a dictonary of params or keyword arguments is correct given a parameter list

    Checks that neccessary parameters exist, and that their type can be casted corretly. There's special logic for list and dictonary types.
    JSON arguments are parsed as dict types.

    Args:
        given_params (dict): keyword arguments to check types
        params_list (List[tuple]): param list, where each element is the parameter (param, type)

    Returns:
        dict: Validated and casted keyword argument dictonary
    """
    logger.info("Validating params...")

    d_params = {}
    for param, dtype in params_list:
        if given_params.get(param, None) is None:
            raise RuntimeError(
                f"Param {param} of type {dtype} was never set, but required by transform, please check your input variables."
            )
        try:
            if "List" in str(dtype):
                if type(given_params[param]) == list:
                    d_params[param] = given_params[param]
                else:
                    d_params[param] = [
                        dtype.__args__[0](s) for s in given_params[param].split(",")
                    ]
            elif dtype == dict:
                if type(given_params[param]) == dict:
                    d_params[param] = given_params[param]
                else:
                    d_params[param] = eval(str(given_params[param]))
            elif type(dtype) == type:
                d_params[param] = dtype(given_params[param])
            else:
                raise RuntimeError(f"Type {type(dtype)} invalid!")

        except ValueError as exc:
            raise RuntimeError(f"Param {param} could not be cast to {dtype} - {exc}")

        except RuntimeError as e:
            raise e
        
        if param in MASK_KEYS:
            logger.info(f" -> Set {param} ({dtype}) = *****")
        else:
            logger.info(f" -> Set {param} ({dtype}) = {d_params[param]}")

    return d_params


def expand_inputs(given_params: dict):
    """For special input_* parameters, see if we should infer the input given an output/result directory

    Args:
        given_params (dict): keyword arguments to check types

    Returns:
        dict: Input- expanded keyword argument dictonary
    """
    d_params = {}
    d_keys = {}

    logger.info("Expanding inputs...")

    for param, param_value in given_params.items():
        if "input_" in param:  # We want to treat input_ params a bit differently

            # For some inputs, they may be defined as a directory, where metadata about them is at the provided directory path
            expected_metadata = os.path.join(param_value, "metadata.yml")
            logger.info(f"Attempting to read metadata at {expected_metadata}")

            if os.path.isdir(param_value) and os.path.exists(
                expected_metadata
            ):  # Check for this metadata file
                # We supplied an inferred input from some previous output, so figure it out from the metadata of this output fold

                with open(expected_metadata, "r") as yaml_file:
                    metadata = yaml.safe_load(yaml_file)

                # Output names/slots are same as input names/slots, just without input_ prefix
                input_type = param.replace("input_", "")

                # Alias flag
                alias = input_type in TYPE_ALIASES.keys()

                # Convert any known type aliases
                input_type = TYPE_ALIASES.get(input_type, input_type)

                # Query the metadata dictionary for that type
                expanded_input = metadata.get(input_type, None)

                # Tree output_directories should never be passed to functions which cannot accept them
                if expanded_input is None:
                    raise RuntimeError(
                        f"No matching output slot of type [{param.replace('input_', '')}] at given input directory"
                    )

                logger.info(f"Expanded input:\n -> {param_value}\n -> {expanded_input}")
                d_params[param] = expanded_input

                # Only propagate keys from non-aliases
                if not alias: 
                    # Query any keys:
                    segment_keys = metadata.get('segment_keys', {})
                    logger.info(f"Found segment keys: {segment_keys}")

                    for key in segment_keys.keys():
                        if key in d_keys.keys() and not segment_keys[key] == d_keys[key]:
                            raise RuntimeError(
                                f"Key mismatch for '{key}', found {segment_keys[key]} and {d_keys[key]}, cannot resolve!!"
                            )
                    
                    d_keys.update(segment_keys)

            else:
                d_params[param] = param_value
        else:
            d_params[param] = param_value

    return d_params, d_keys



def get_dataset_url():
    """ Retrieve a "dataset URL" from the environment, may look like http://localhost:6077 or file:///absolute/path/to/dataset/dir """
    dataset_url = os.environ.get("DATASET_URL", None)

    if dataset_url is None:
        logger.warning("Requesting feature data be sent to dataset, however no dataset URL provided, please set env DATASET_URL!")
    else:
        logger.info(f"Found dataset URL = {dataset_url}")

    return dataset_url

    

def post_to_dataset(input_feature_data, waystation_url, dataset_id, keys):
    """ Interface feature data to a parquet dataset

    Args:
        input_feature_data (str): path to input data
        waystation_url (str): URL of dataset root (either file or using waystation)
        dataset_id (str): Dataset name/ID
        keys (dict): corresponding segment keys
    """

    logger.info(f"Adding {input_feature_data} to {dataset_id} via {waystation_url}")

    segment_id = "-".join(
        [v for _, v in sorted(keys.items())]
    )

    logger.info(f"SEGMENT_ID={segment_id}")
    
    post_url = os.path.join ( waystation_url, "datasets", dataset_id, "segments", segment_id )

    parsed_url = urllib.parse.urlparse(post_url)

    if 'http' in parsed_url.scheme:
        # The cool way, using luna waystation

        logger.info (f"Posting to: {post_url}")

        res = requests.post(post_url, files={'segment_data': open (input_feature_data, 'rb')}, data={"segment_keys": json.dumps(keys)})

        logger.info (f"{res}: {res.text}")

    elif 'file' in parsed_url.scheme:
        # The less cool way, just using file paths

        segment_dir = Path ( parsed_url.path )

        logger.info (f"Writing to: {segment_dir}")

        os.makedirs(segment_dir, exist_ok=True)

        shutil.copy(input_feature_data, segment_dir.joinpath("data.parquet"))
    
    else:
        logger.warning("Unrecognized scheme: {parsed_url.scheme}, skipping!")



def cli_runner(
    cli_kwargs: dict, cli_params: List[tuple], cli_function: Callable[..., dict], pass_keys: bool = False
):
    """For special input_* parameters, see if we should infer the input given an output/result directory

    Args:
        cli_kwargs (dict): keyword arguments from the CLI call
        cli_params (List[tuple]): param list, where each element is the parameter (name, type)
        cli_function (Callable[..., dict]): cli_function entry point, should accept exactly the arguments given by cli_params
        pass_keys (bool): will pass found segment keys to transform function as 'keys' kwarg

    Returns:
        None

    """
    logger.info(f"Started CLI Runner wtih {cli_function}")
    logger.debug(f"cli_kwargs={cli_kwargs}")
    logger.debug(f"cli_params={cli_params}")
    logger.debug(f"pass_keys={pass_keys}")

    trm_kwargs = {}

    # if "output_dir" not in cli_kwargs.keys():
    #    raise RuntimeError("CLI Runners assume an output directory")

    # Get params from param file
    if cli_kwargs.get("method_param_path"):
        with open(cli_kwargs.get("method_param_path"), "r") as yaml_file:
            yaml_kwargs = yaml.safe_load(yaml_file)
        trm_kwargs.update(yaml_kwargs)  # Fill from json

    for key in list(cli_kwargs.keys()):
        if cli_kwargs[key] is None:
            del cli_kwargs[key]

    # Override with CLI arguments
    trm_kwargs.update(cli_kwargs)

    trm_kwargs = validate_params(trm_kwargs, cli_params)

    if "output_dir" in trm_kwargs:
        output_dir = trm_kwargs["output_dir"]
        os.makedirs(output_dir, exist_ok=True)

    # Expand implied inputs
    trm_kwargs, keys = expand_inputs(trm_kwargs)

    logger.info (f"Full segment key set: {keys}")

    # Nice little log break
    logger.info(
          "-" * 60
        + f"\n Starting transform::{cli_function.__name__} \n"
        + "-" * 60
    )

    with CodeTimer(logger, name=f"transform::{cli_function.__name__}"):
        if pass_keys: cli_function = partial (cli_function, keys=keys)

        result = cli_function(**trm_kwargs)

    # Nice little log break
    logger.info(
          "-" * 60
        + f"\n Done with transform, running post-transform functions... \n"
        + "-" * 60
    )

    trm_kwargs.update(result)

    # filter out kwargs with sensitive data
    for key in MASK_KEYS:
        trm_kwargs.pop(key, None)
    
    # propagate keys
    if trm_kwargs.get('segment_keys', None):
        trm_kwargs['segment_keys'].update(keys)
    else: 
        trm_kwargs['segment_keys'] = keys

    # Save metadata on disk
    if "output_dir" in trm_kwargs:
        with open(os.path.join(output_dir, "metadata.yml"), "w") as fp:
            yaml.dump(trm_kwargs, fp)

    # Save feature data in parquet if indicated:
    if "dataset_id" in cli_kwargs and  "feature_data" in trm_kwargs:
        dataset_id   = cli_kwargs.get("dataset_id")
        feature_data = trm_kwargs.get("feature_data")

        logger.info(f"Adding feature segment {feature_data} to {dataset_id}")

        dataset_url = get_dataset_url()

        if dataset_url is not None:
            post_to_dataset( feature_data, dataset_url, dataset_id, keys=trm_kwargs['segment_keys'])


    logger.info("Done.")


def apply_csv_filter(input_paths, subset_csv=None):
    """Filteres a list of input_paths based on include/exclude logic given for either the full path, filename, or filestem

    If using "include" logic, only matching entries with include=True are kept.
    If using "exclude" logic, only matching entries with exclude=True are removed.

    The origional list is returned if the given subset_csv is None or empty

    Args:
        input_paths (list[str]): list of input paths to filter
        subset_csv (str): path to a csv with subset/filter information/flags
    Returns
        list[str]: filtered list
    Raises:
        RuntimeError: If the given subset_csv is invalid
    """

    if not len(subset_csv) > 0 or subset_csv is None:
        return input_paths
    if not os.path.exists(subset_csv):
        return input_paths

    try:
        subset_df = pd.read_csv(subset_csv, dtype={0: str})

        match_type = subset_df.columns[0]
        filter_logic = subset_df.columns[1]

        if match_type not in ["path", "filename", "stem"]:
            raise RuntimeError("Invalid match type column")
        if filter_logic not in ["include", "exclude"]:
            raise RuntimeError("Invalid match type column")
    except Exception as exc:
        logger.error(exc)
        raise RuntimeError(
            "Invalid subset .csv passed, must be a 2-column csv with headers = [ (path|filename|stem), (include|exclude) ]"
        )

    if not len(subset_df) > 0:
        return input_paths

    logger.info(
        f"Applying csv filter, match_type={match_type}, filter_logic={filter_logic}"
    )

    input_path_df = pd.DataFrame(
        {"path": path, "filename": Path(path).name, "stem": Path(path).stem}
        for path in input_paths
    ).astype(str)

    df_matches = input_path_df.set_index(match_type).join(
        subset_df.set_index(match_type)
    )

    if filter_logic == "include":
        out = df_matches.loc[(df_matches["include"] == 1)]
    if filter_logic == "exclude":
        out = df_matches.loc[df_matches["exclude"] == 1]

    return list(out.reset_index()["path"])


def load_func(dotpath: str):
    """load function in module from a parsed yaml string

    Args:
        dotpath (str): module/function name written as a string (ie torchvision.models.resnet34)
    Returns:
        The inferred module itself, not the string representation
    """
    module_, func = dotpath.rsplit(".", maxsplit=1)
    m = import_module(module_)
    return getattr(m, func)



from pyspark.sql import SparkSession
from luna.common.config import ConfigSet
import warnings
"""Common spark session"""
class SparkConfig:

	def spark_session(self, config_name, app_name):
		'''
		@:param config_name logical name of the configuration to use from the ConfigSet. See luna/common/config.py
		@:param app_name application name to give to the spark session. This name will be used in the spark logs.
		'''

		cfg = ConfigSet()
		warnings.warn("You are using SparkConfig to generate a spark_session, however spark has been depreciated from this package!")

		spark_uri = cfg.get_value(path=config_name+'::$.spark_cluster_config[:1]["spark.uri"]')
		spark_driver_host = cfg.get_value(path=config_name+'::$.spark_cluster_config[:2]["spark.driver.host"]')
		spark_executor_cores = cfg.get_value(path=config_name+
												  '::$.spark_application_config[:1]["spark.executor.cores"]')
		spark_cores_max = cfg.get_value(path=config_name+'::$.spark_application_config[:2]["spark.cores.max"]')
		spark_executor_memory = cfg.get_value(path=config_name+
												   '::$.spark_application_config[:3]["spark.executor.memory"]')
		spark_executor_pyspark_memory = \
			cfg.get_value(path=config_name+'::$.spark_application_config[:4]["spark.executor.pyspark.memory"]')
		spark_sql_shuffle_partitions = \
			cfg.get_value(path=config_name+'::$.spark_application_config[:5]["spark.sql.shuffle.partitions"]')
		spark_driver_maxresultsize = \
			cfg.get_value(path=config_name+'::$.spark_application_config[:6]["spark.driver.maxResultSize"]')

		return SparkSession.builder \
			.appName(app_name) \
			.master(spark_uri) \
			.config("spark.jars.packages", "io.delta:delta-core_2.12:0.7.0") \
			.config("spark.delta.logStore.class", "org.apache.spark.sql.delta.storage.HDFSLogStore") \
			.config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension") \
			.config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog") \
			.config("spark.hadoop.dfs.client.use.datanode.hostname", "true") \
			.config("spark.driver.host", spark_driver_host) \
			.config("spark.sql.execution.arrow.pyspark.enabled", "true") \
			.config("spark.executor.memory", spark_executor_memory) \
			.config("spark.driver.memory", spark_executor_memory) \
			.config("spark.executor.cores", spark_executor_cores) \
			.config("spark.cores.max", spark_cores_max) \
			.config("spark.executor.pyspark.memory", spark_executor_pyspark_memory) \
			.config("spark.sql.shuffle.partitions", spark_sql_shuffle_partitions) \
			.config("spark.driver.maxResultSize", spark_driver_maxresultsize) \
			.config("spark.files.overwrite", "true") \
			.config("fs.defaultFS", "file:///") \
			.config("spark.driver.extraJavaOptions", "-Dio.netty.tryReflectionSetAccessible=true") \
			.config("spark.executor.extraJavaOptions", "-Dio.netty.tryReflectionSetAccessible=true") \
			.getOrCreate()




'''
Created on October 17, 2019

@author: pashaa@mskcc.org
'''

import yaml
import yamale
import logging
from jsonpath_ng import parse
from luna.common.utils import get_absolute_path

logger = logging.getLogger(__name__)

class ConfigSet():
    '''
    This is a singleton class that can load a collection of configurations from yaml files.

    ConfigSet loads configurations from yaml files only once on first invocation of this class with
    the specified yaml file. The class then maintains the configuration in memory in a singleton instance. All new
    invocations of this class will serve up the same configuration.

    Each configuration in the collection is identified by a logical name.

    If a new invocation of this class is created with an existing logical name and a different yaml file, the singleton
    instance replaces the existing configuration with the newly specified yaml file for the given logical name.
    '''

    __CONFIG_MAP = {}  # maps logical name to yaml config file name
    __SCHEMA_MAP = {}  # maps logical name to schema file of yaml config file
    __INSTANCE = None    # singleton instance containing the collection of configs keyed by logical name


    def __new__(cls, name=None, config_file=None, schema_file=None):
        # assume one or more collections have already been loaded
        if name is None or config_file is None:
            return ConfigSet.__INSTANCE

        # initialize singleton
        if ConfigSet.__INSTANCE is None:
            ConfigSet.__INSTANCE = object.__new__(cls)
            ConfigSet.__INSTANCE.__config = {}

        # load or reload config into memory
        if name not in ConfigSet.__CONFIG_MAP.keys() or \
            ConfigSet.__CONFIG_MAP[name] != config_file:
            ConfigSet.__CONFIG_MAP[name] = config_file
            ConfigSet.__INSTANCE.__config[name] = ConfigSet._load_config(cls, name)

            # add schema and validate config
            if schema_file is not None:
                ConfigSet.__SCHEMA_MAP[name] = schema_file
                ConfigSet._validate_config(cls, name)

        return ConfigSet.__INSTANCE


    def __init__(self, name=None, config_file=None, schema_file=None):
        '''
        :param name logical name to be given for this configuration. This argument only needs to be provided on first
                    invocation (optional).
        :param config_file the config file to load. This argument only needs to be provided on first
                           invocation (optional).
        :param schema_file a schema file for the yaml configuration (optional)
        :raises yamale.yamale_error.YamaleError if config file is invalid when validated against the schema
        '''
        pass  # see __new__() method implementation

    def _validate_config(cls, name):
        config_file = ConfigSet.__CONFIG_MAP[name]
        schema_file = ConfigSet.__SCHEMA_MAP[name]
        logger.info("validating config " + config_file + " against schema " + schema_file + " for " + name)
        schema = yamale.make_schema(schema_file)
        data = yamale.make_data(config_file)
        yamale.validate(schema, data)

    def _load_config(cls, name):
        '''

        :param name: logical name of the config to load
        :return: config generator object

        :raises: IOError if yaml config file for the specified logical name cannot be found
        '''
        # read config file
        config_file = ConfigSet.__CONFIG_MAP[name]
        logger.info("loading config file "+config_file)


        try:
            stream = open(config_file, 'r')
        except IOError as err:
            logger.error("unable to find a config file with name "+config_file+
                  ". Please use config.yaml.template to make a "+config_file+". "+str(err))
            raise err

        config = {}
        for items in yaml.load_all(stream, Loader=yaml.FullLoader):
            config.update(items)

        return config

    def _parse_path(self, path):
        path_segments = path.split('::', 1)  # split just once

        if len(path_segments) != 2:
            err = 'Illegal config path: '+path+'. must be of form "name::jsonpath" ' \
                        'where name is the logical name of the configuration and jsonpath is the ' \
                                                      'jsonpath into the yaml configuration'
            logger.error(err)
            raise ValueError(err)

        return {'name': path_segments[0], 'jsonpath': path_segments[1]}


    def _get_match(self, name, jsonpath):
        jsonpath_expression = parse(jsonpath)

        return jsonpath_expression.find(ConfigSet.__INSTANCE.__config[name])


    def has_value(self, path):
        '''
        Args:
            path (str): path to a value in a configuration. The path must be of the form
            "name::jsonpath" where name is the logical name of the configuration and jsonpath is the jsonpath to value.
            see config.yaml to generate a jsonpath. See https://pypi.org/project/jsonpath-ng/ jsonpath expressions
            may be tested here - https://jsonpath.com/

        Returns:
            boolean: true if value is not an empty string, else false.

        Raises:
            ValueError: if a configuration with the specified name was never loaded

        '''
        parsed = self._parse_path(path)
        name= parsed['name']
        jsonpath = parsed['jsonpath']

        if ConfigSet.__INSTANCE is None or name not in ConfigSet.__INSTANCE.__config.keys():
            raise ValueError('configuration with logical name '+name+' was never loaded')

        if len(self._get_match(name, jsonpath)) == 0:
            return False
        else:
            return True


    def get_value(self, path):
        '''
        Gets the value for the specified jsonpath from the specified configuration.

        Args:
            path (str): path to a value in a configuration. The path must be of the form "name::jsonpath"
            where name is the logical name of the configuration and jsonpath is the jsonpath to value.
            see config.yaml to generate a jsonpath. See https://pypi.org/project/jsonpath-ng/
            jsonpath expressions may be tested here - https://jsonpath.com/

        Returns:
            str: value from config file

        Raises:
            ValueError: if no match is found for the specified exception or a configuration with
            the specified name was never loaded

        '''
        parsed = self._parse_path(path)
        name = parsed['name']
        jsonpath = parsed['jsonpath']

        if ConfigSet.__INSTANCE is None or name not in ConfigSet.__INSTANCE.__config.keys():
            raise ValueError('configuration with logical name '+name+' was never loaded')

        match = self._get_match(name, jsonpath)

        if len(match) == 0:
            err = 'unable to find a config value for jsonpath: '+jsonpath
            logger.error(err)
            raise ValueError(err)

        return match[0].value


    def get_names(self):
        '''

        :return: a list of logical names of the configs stored in this instance.
        '''
        if ConfigSet.__INSTANCE is not None:
            return list(ConfigSet.__INSTANCE.__config.keys())
        else:
            return []


    def get_keys(self, name):
        '''

        :param name: logical name of the configuration
        :return: a list of top-level keys in the config stored in this instance.
        :raises: ValueError if a configuration with the specified name was never loaded
        '''
        if ConfigSet.__INSTANCE is None or name not in ConfigSet.__INSTANCE.__config.keys():
            raise ValueError('configuration with logical name '+name+' was never loaded')

        return list(ConfigSet.__INSTANCE.__config[name].keys())

    def get_config_set(self, name):
        '''

        :param name: logical name of the configuration
        :return: a dictonary of top-level keys in the config stored in this instance.
        :raises: ValueError if a configuration with the specified name was never loaded

        '''
        if ConfigSet.__INSTANCE is None or name not in ConfigSet.__INSTANCE.__config.keys():
            raise ValueError('configuration with logical name '+name+' was never loaded')
        return ConfigSet.__INSTANCE.__config[name]



    def clear(self):
        '''
        clear the entire collection of configurations
        '''
        ConfigSet.__CONFIG_MAP = {}
        ConfigSet.__SCHEMA_MAP = {}
        ConfigSet.__INSTANCE = None


if __name__ == '__main__':
    c1 = ConfigSet('app_config', 'pyluna-pathology/tests/luna/common/test_config.yml')
    c2 = ConfigSet('app_config', 'pyluna-pathology/tests/luna/common/test_config.yml')
    c3 = ConfigSet('app_config', 'pyluna-pathology/tests/luna/common/test_config.yml')

    print(str(c1) + ' ' + str(c1.get_value('app_config::$.spark_application_config[:1]["spark.executor.cores"]')))
    print(str(c2) + ' ' + str(c2.get_value('app_config::$.spark_application_config[:1]["spark.executor.cores"]')))
    try:
        print(str(c3) + ' ' + str(c3.get_value('app_config::$.spark_application_config[:1]["doesnt_exist"]')))
    except ValueError as ve:
        print("got expected value error: "+str(ve))

    schema_file = get_absolute_path(__file__, '../data_ingestion_template_schema.yml')

    c4 = ConfigSet(name='data_config',
                   config_file="pyluna-pathology/tests/luna/common/test_data_ingestion_template.yml",
                   schema_file=schema_file)

    print(str(c4) + ' ' + str(c4.get_value('data_config::$.REQUESTOR')))



from luna.common.utils import to_sql_field, to_sql_value, does_not_contain
import warnings, os
from pathlib import Path

# TODO: Update types to match docs
CONTAINER_TYPES = ["cohort", "patient", "scan", "slide", "parquet", "accession", "generic"]
RADIOLOGY_TYPES = ["DicomSeries", "DicomImageSeries", "DicomImage", "VolumetricImage", "RadiologyScan", "VolumetricLabel", "VolumetricLabelSet", "Voxels", "Radiomics"]
PATHOLOGY_TYPES = ["WholeSlideImage", "WsiThumbnail", "TileScores", "TileImages", "PointAnnotation", "PointAnnotationJson", "RegionalAnnotationBitmap", "RegionalAnnotationJSON", "CellMap"]
GENERIC_TYPES   = ["ResultSegment"]
ALL_DATA_TYPES  = RADIOLOGY_TYPES + PATHOLOGY_TYPES + GENERIC_TYPES

class Node(object):
	"""
	Node object defines the type and attributes of a graph node.

	:param: node_type: node type. e.g. scan
	:param: name: required node name. e.g. scan-123
	:param: properties: dict of key, value pairs for the node.
	"""
	def __init__(self, node_type, node_name, properties=None):

		# Required schema: node_type, node_name

		self.type = node_type

		if properties is None:
			self.properties = {}
		else:
			self.properties = properties

		# Special case: a cohort name is it's own namespace
		if self.type in ["cohort"]:
			self.properties['namespace'] = node_name

		# For containers, DB name is the name, for data type nodes, it's type-name, and must be one of these two categories
		if self.type in CONTAINER_TYPES:
			self.name = f'{node_name}'
		elif self.type in ALL_DATA_TYPES:
			self.name = f'{node_type}-{node_name}'
		else:
			raise RuntimeError(f"Invalid Node Data Type {self.type}")


		if   "namespace" in self.properties.keys() and "subspace" in self.properties.keys():
                        self.properties["qualified_address"] = self.get_qualified_name(self.properties['namespace'], self.properties['subspace'], self.name)
		elif "namespace" in self.properties.keys():
			self.properties["qualified_address"] = self.get_qualified_name(self.properties['namespace'], self.name)
		else:
			self.properties['qualified_address'] = self.name

		self.properties["type"] = self.type

		self.set_data( self.properties.get("data", None))
		self.set_aux ( self.properties.get("aux",  None))

	def set_namespace(self, namespace_id: str, subspace_id=None):
		"""
		Sets the namespace for this Node commits

		:params: namespace_id - namespace value
		:params: subspace_id  - subspace value, optional
		"""
		self.properties['namespace'] = namespace_id

		if subspace_id is None:
			self.properties["qualified_address"] = self.get_qualified_name(self.properties['namespace'], self.name)
		else:
			self.properties['subspace']  = subspace_id
			self.properties["qualified_address"] = self.get_qualified_name(self.properties['namespace'], self.properties['subspace'], self.name)


	def set_data(self, data):
		self.data = None

		if data is None: return

		if isinstance(data, str):
			data = Path(data)

		if not isinstance(data, Path):
			raise RuntimeError("set_data() only accepts path-like objects")

		if not data.exists():
			raise RuntimeError("Tried to set data to a non-existent path!")

		self.properties['data'] = str(data)
		self.data = str(data)

	def set_aux(self, aux):
		self.aux = None

		if aux is None: return

		if isinstance(aux, str):
			aux = Path(aux)

		if not isinstance(aux, Path):
			raise RuntimeError("set_aux() only accepts path-like objects")

		if not aux.exists():
			raise RuntimeError("Tried to set aux to a non-existent path!")

		self.properties['aux'] = str(aux)
		self.aux = str(aux)

	def __repr__(self):
		"""
		Returns a string representation of the node
		"""
		kv = self.get_all_props()
		prop_string = self.prop_str_repr(kv.keys(), kv)
		bigline = "-"*100
		return f"{bigline}\nname: {self.name}\ntype: {self.type}\nproperties: \n{prop_string}\n{bigline}"

	def get_all_props(self):
		"""
		Name is a required field, but it's still a property of this node.
		Return the properties as a dict including the name property!
		"""
		kv = self.properties
		kv["name"] = self.name

		return kv

	def get_create_str(self):
		"""
		Returns a string representation of the node with all properties
		"""
		kv = self.get_all_props()

		prop_string = self.prop_str(kv.keys(), kv)
		return f"""{self.type}:globals{{ {prop_string} }}"""

	def get_match_str(self):
		"""
		Returns a string representation of the node with only the qualified_address as a property
		"""
		kv = self.get_all_props()

		prop_string = self.prop_str( ["qualified_address"], kv)
		return f"""{self.type}:globals{{ {prop_string} }}"""

	def get_map_str(self):
		"""
		Returns the properties as a cypher map
		"""
		kv = self.get_all_props()

		prop_string = self.prop_str(kv.keys(), kv)
		return f"""{{ {prop_string} }}"""

	def get_address(self):
		"""
		Returns current node address
		"""
		return self.properties.get("qualified_address")

	@staticmethod
	def prop_str(fields, row):
		"""
		Returns a kv string like 'id: 123, ...' where prop values come from row.
		"""
		fields = set(fields).intersection(set(row.keys()))

		kv = [f" {to_sql_field(x)}: {to_sql_value(row[x])}" for x in fields]
		return ','.join(kv)

	@staticmethod
	def prop_str_repr(fields, row):
		"""
		Returns a kv string like ' - id: 123 <newline> ...' where prop values come from row.
		"""
		fields = set(fields).intersection(set(row.keys()))

		kv = [f"   {to_sql_field(x)}: {to_sql_value(row[x])}" for x in fields]
		return '\n'.join(kv)

	@staticmethod
	def get_qualified_name(*args):
		"""
		Returns the full name given a namespace and patient ID
		"""
		for name in args: does_not_contain(":", name)

		return "::".join(args)




import os
import pytest

from luna.common.DataStore import DataStore_v2

@pytest.fixture()
def datastore2():
    os.environ['LUNA_HOME'] = 'pyluna-common/tests/luna/common/testdata/'
    datastore2 = DataStore_v2('pyluna-common/tests/luna/common/testdata/store')
    yield datastore2

def test_datastore_v2_init():
    os.environ['LUNA_HOME'] = ''
    with pytest.raises(RuntimeError):
        DataStore_v2('pyluna-common/tests/luna/common/testdata/store')



from luna.common.Neo4jConnection import Neo4jConnection
from luna.common.Node import Node, CONTAINER_TYPES
from luna.common.config import ConfigSet

import os, socket, pathlib, logging, shutil
from minio import Minio

from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed

logger = logging.getLogger(__name__)

class DataStore_v2:
    def __init__(self, store_location):
        if os.environ['LUNA_HOME']:
            self.params = ConfigSet(name='STORE_CFG',
                                    config_file=os.path.join(os.environ['LUNA_HOME'], 'conf', 'datastore.cfg')) \
                .get_config_set("STORE_CFG")
        else:
            raise RuntimeError(
                "$LUNA_HOME is not set. Make sure you have set $LUNA_HOME and $LUNA_HOME/conf/datastore.cfg")
        logger.info(f"Configured datastore with {self.params}")

        self.backend = store_location
        os.makedirs(self.backend, exist_ok=True)
        logger.info(f"Datstore file backend= {self.backend}")

    def ensure_datastore(self, datastore_id, datastore_type):
        """
        :params: datastore_id - unique container ID
        :params: datastore_type - the type of the container

        """
        datastore_id = str(datastore_id)

        if not datastore_type in CONTAINER_TYPES:
            logger.warning (f"DataStore type [{datastore_type}] invalid, please choose from [{CONTAINER_TYPES}]" )
            return

        if ":" in datastore_id:
            logger.warning (f"Invalid datastore_id [{datastore_id}], only use alphanumeric characters")
            return

        conn = Neo4jConnection(uri=self.params['GRAPH_URI'], user=self.params['GRAPH_USER'], pwd=self.params['GRAPH_PASSWORD'])
        res = conn.query(f""" MERGE (datastore:globals:{datastore_type}{{qualified_address:'{datastore_id}'}}) RETURN count(datastore)""")

        if res[0]['count(datastore)']==1:
            logger.info(f"DataStore [{datastore_id}] of type [{datastore_type}] created or matched successfully!")
        else:
            logger.error("The datastore {node} could not be created or found")

    def _write_to_graph_store(self, node, store_id):
        """ Saves the 'node' to a datastore managed in the graph DB """

        try:
            # Configure our connection
            conn = Neo4jConnection(uri=self.params['GRAPH_URI'], user=self.params['GRAPH_USER'], pwd=self.params['GRAPH_PASSWORD'])
            res = conn.query( f"""
                MATCH (datastore) WHERE datastore.qualified_address = '{store_id}'
                MERGE (datastore)-[:HAS_DATA]->(da:{node.get_match_str()})
                    ON MATCH  SET da = {node.get_map_str()}
                    ON CREATE SET da = {node.get_map_str()}
                RETURN count(datastore)""" )
            if res is None:
                logger.error(f"Tried adding data to {store_id}, however query failed, this data will not be available!", extra={'store_id': store_id})
                return
            if not res[0]['count(datastore)']==1:
                logger.warning(f"Tried adding data to {store_id}, however datastore did not exist, this data will not be available!", extra={'store_id': store_id})
                return
        except Exception as exc:
            logger.exception(f"On write, encountered {exc}, continuing...", extra={'store_id': store_id})

    def get(self, store_id, namespace_id, data_type, data_tag='data', realpath=True):
        """ Looks up and returns the path of data given the store_id, namespace_id, data_type, and data_tag """

        dest_dir = os.path.join (self.backend, store_id, namespace_id, data_type, data_tag)
        if not os.path.exists(dest_dir):
            # if realpath is true, return path to data instead of symlink location
            if os.path.lexists(dest_dir):
                if realpath:
                    dest_dir = os.readlink(dest_dir)
            else:
                raise RuntimeWarning(f"Data not found at {dest_dir}")
        return dest_dir

    def put(self, filepath, store_id, namespace_id, data_type, data_tag='data', metadata={}, symlink=False):
        """ Puts the file at filepath at the proper location given a store_id, namespace_id, data_type, and data_tag, and save metadata to DB """

        dest_dir = os.path.join (self.backend, store_id, namespace_id, data_type, data_tag)

        if symlink:
            os.makedirs(pathlib.Path(dest_dir).parent, exist_ok=True)

            if os.path.lexists(dest_dir):
                os.remove(dest_dir)
            logger.info(f"Create symlink {dest_dir} -> {filepath}")
            os.symlink(filepath, dest_dir)
        else:
            os.makedirs(dest_dir, exist_ok=True)
            logger.info(f"Save {filepath} -> {dest_dir}")
            shutil.copy(filepath, dest_dir )

        if self.params['GRAPH_STORE_ENABLED']:
            node = Node(data_type, data_tag, metadata)
            node.set_namespace(namespace_id, store_id)
            logger.info(f"Adding: {node}")
            self._write_to_graph_store (node, store_id)

        return dest_dir

    def write(self, iostream, store_id, namespace_id, data_type, data_tag, metadata={}, dtype='w'):
        """ Writes iostream at the proper location given a store_id, namespace_id, data_type, and data_tag, and save metadata to DB """

        dest_path_dir  = os.path.join (store_id, namespace_id, data_type)
        dest_path_file = os.path.join (dest_path_dir, data_tag)

        dest_dir  = os.path.join (self.backend, dest_path_dir)
        dest_file = os.path.join (self.backend, dest_path_file)

        os.makedirs(dest_dir, exist_ok=True)
        logger.info(f"Save -> {dest_file}")
        with open(dest_file, dtype) as fp:
            fp.write(iostream)

        if self.params['GRAPH_STORE_ENABLED']:
            node = Node(data_type, data_tag, metadata)
            node.set_namespace(namespace_id, store_id)
            logger.info(f"Adding: {node}")
            self._write_to_graph_store (node, store_id)

        return dest_file



def bootstrap (container_id):
    logger.info(f"Bootstrapping pipeline for {container_id}")
    return 1

class DataStore(object):
    """
    DataStore: an abstraction with an id, name, namespace, type, and a list of associated data nodes

    Interfaces with a metadata store (graph DB) and raw file stores (gpfs, potentially others)

    Handles the matching and creation of metadata

    Example usage:

    $ container = luna.common.GraphEnum.DataStore( params ).setNamespace("test").setContainer("1.2.840...")
    > Connecting to: neo4j://localhost:7687
    > Connection successfull: True
    > Running on: localhost
    > Lookup ID: 1.2.840...
    > Found: [<Record id(container)=7091 labels(container)=['scan'] container.type='scan' container.name='1.2.840...>]
    > Match on: WHERE id(container) = 7091
    > Successfully attached to: scan 1.2.840...

    $ node = Node("dicom", "DCM-0123", {"Modality":"CT", "path":"file:/some/path/1.dcm"})

    $ container.put(node)
    > Adding: test-0000
    DataStore has 1 pending commits

    $
    > Committing dicom:globals{ hash: 'abc123' name: 'my-dicom', qualified_address: 'test::1.2.840::my-dicom', namespace: 'test', type: 'dicom' , path: 'file:/some/path/1.dcm'}

    $ container.get("dicom", "my-dicom").path
    > /some/path/1.dcm

    $ container.get("dicom", "my-dicom").properties['Modality']
    > 'CT'

    The container includes a logging method:
    $ container.logger.info("I am processing the CT")
    > 'yyyy-mm-dd h:m:s,ms - DataStore [1] - INFO - I am processing the CT'

    """
    # TODO: worried about schema issues? like making sure name, namespace, type and qualified path are present, neo4j offers schema enforcment.
    # TODO: testing
    # TODO: error checking

    def __init__(self, params):
        """
        Initialize the container object.
            Connects to the graph DB
            Figure out what host this code is running on

        :params: params - dictonary of important configuration, right now, only the graph URI connection parameters are needed.
        """
        if isinstance(params, ConfigSet):
            params=params.get_config_set("APP_CFG")

        # Connect to graph DB
        logger.debug ("Connecting to: %s", params['GRAPH_URI'])
        self._conn = Neo4jConnection(uri=params['GRAPH_URI'], user=params['GRAPH_USER'], pwd=params['GRAPH_PASSWORD'])
        logger.debug ("Connection test: %s", self._conn.test_connection())

        if params.get('OBJECT_STORE_ENABLED',  False):
            logger.debug ("Connecting to: %s", params['MINIO_URI'])
            self._client = Minio(params['MINIO_URI'], access_key=params['MINIO_USER'], secret_key=params['MINIO_PASSWORD'], secure=False)
            try:
                for bucket in self._client.list_buckets():
                    logger.debug("Found bucket %s", bucket.name )
                logger.debug("OBJECT_STORE_ENABLED=True")
                params['OBJECT_STORE_ENABLED'] = True
            except:
                logger.warning("Could not connect to object store")
                logger.warning("Set OBJECT_STORE_ENABLED=False")
                params['OBJECT_STORE_ENABLED'] = False

        self._host = socket.gethostname() # portable to *docker* containers
        logger.debug ("Running on: %s", self._host)

        self.params = params
        self._attached = False

    def createNamespace(self, namespace_id: str):
        """
        Creates a namesapce, if it doesn't exist, else, tells you it exists

        :params: namespace_id - namespace value
        """
        cohort = Node("cohort", namespace_id)
        create_res = self._conn.query(f""" MERGE (co:{cohort.get_create_str()}) RETURN co""")

        if len(create_res) == 1:
            logger.info(f"Namespace [{namespace_id}] created successfully")

        return self

    def setNamespace(self, namespace_id: str):
        """
        Sets the namespace for this container's commits, if it exists

        :params: namespace_id - namespace value
        """
        self._namespace_id   = namespace_id
        self._namespace_node = Node("cohort", namespace_id)
        self._bucket_id      = namespace_id.lower().replace('_','-')

        logger.debug(f"Checking if [{namespace_id}] exists...")

        match_res = self._conn.query(f""" MATCH (co:{self._namespace_node.get_match_str()}) RETURN co""")

        if not len(match_res) == 1:
            raise RuntimeError( f"Namespace [{namespace_id}] does not exist, call .createNamespace() first!")

        if self.params.get('OBJECT_STORE_ENABLED',  False):
            if not self._client.bucket_exists(self._bucket_id):
                self._client.make_bucket(self._bucket_id)

        return self

    def createDatastore(self, container_id, container_type):
        """
        Checks if the node referenced by container_id is a valid container, queries the metastore for relevant metadata

        :params: container_id - unique container ID
        :params: type - the type of the container

        """

        if not container_type in ['generic', 'patient', 'accession', 'scan', 'slide', 'parquet']:
            logger.warning (f"DataStore type [{container_type}] invalid, please choose from ['generic', 'patient', 'accession', 'scan', 'slide', 'parquet']" )

        if ":" in container_id:
            logger.warning (f"Invalid container_id [{container_id}], only use alphanumeric characters")

        node = Node(container_type, container_id)
        node.set_namespace( self._namespace_id )

        create_res = self._conn.query(f""" MERGE (container:{node.get_create_str()}) RETURN container""")

        if len(create_res)==1:
            logger.info(f"DataStore [{container_id}] of type [{container_type}] created or matched successfully!")
        else:
            logger.error("The container does not exists")

        return self

    def setDatastore(self, container_id):
        """
        Checks if the node referenced by container_id is a valid datastore, queries the metastore for relevant metadata

        :params: container_id - the unique container ID, either as an integer (neo4j autopopulated ID) or as a string (the Qualified Path)
        """
        self._attached = False
        logger.info ("Lookup ID: %s", container_id)

        # Figure out how to match the node
        if isinstance(container_id, str) and not "uid://" in container_id:
            node = Node("generic", container_id)
            node.set_namespace( self._namespace_id )
            print (node.get_address())
            match_clause = f"""WHERE container.qualified_address = '{node.get_address()}' """
        elif isinstance(container_id, str) and "uid://" in container_id:
            match_clause = f"""WHERE id(container) = {container_id.replace('uid://', '')} """
        elif isinstance(container_id, int):
            match_clause = f"""WHERE id(container) = {container_id} """
        else:
            raise RuntimeError("Invalid container_id type not (str, int)")

        # Run query
        res = self._conn.query(f"""
            MATCH (container) {match_clause}
            RETURN id(container), labels(container), container.type, container.name, container.namespace, container.qualified_address"""
        )

        # Check if the results are singleton (they should be... since we only query unique IDs!!!)
        if res is None or len(res) == 0:
            logger.warning (f"DataStore [{container_id}] does not exist, you can try creating it first with createContainer()")
            return self

        # Set some potentially import parameters
        self._datastore_id  = res[0]["id(container)"]
        self._name          = res[0]["container.name"]
        self._qualifiedpath = res[0]["container.qualified_address"]
        self._type          = res[0]["container.type"]
        self._labels        = res[0]["labels(container)"]
        self.address        = res[0]["container.qualified_address"]

        # Containers need to have a qualified path
        if self._qualifiedpath is None:
            logger.warning ("Found, however not valid container object, containers must have a name, namespace, and qualified path")
            return self

        # Let us know attaching was a success! :)
        logger.info ("Successfully attached to %s container id=%s @ %s", self._type, self._datastore_id, self.address)
        self._attached = True

        return self

    def isAttached(self):
        """
        Returns true if container was properly attached (i.e. checks in setDatastore succeeded), else False
        """
        logger.debug ("Attached: %s", self._attached)
        return self._attached


    def get(self, type, name):
        """
        Query graph DB container node for dependent data nodes, and return one
        Parses the path field URL for various cases, and sets the node.data an node.aux attribute with a corrected path
        Note: namespace is not a default filter for get nodes, but is for adding them (i.e., one can write data under a different namespace)

        :params: type - the type of data designed
            e.g. radiomics, mha, dicom, png, svs, geojson, etc.
        :params: name - can be used to filter nodes
            e.g. name of the node in the subspace of the container (e.g. generate-mhd)
        :example: get("mhd", "generate-mhd") gets data of type "mhd" generated from the method "generate-mhd" in this container's context/subspace
        """
        assert self.isAttached()

        query = f"""MATCH (container)-[:HAS_DATA]-(data:{type}) WHERE id(container) = {self._datastore_id} AND data.name='{type}-{name}' AND data.namespace='{self._namespace_id}' RETURN data"""

        logger.debug(query)
        res = self._conn.query(query)

        # Catches bad queries
        # If successfull query, reconstruct a Node object
        if res is None:
            logger.warning(f"get() query failed, data.name='{type}-{name}' returning None")
            return None
        elif len(res) == 0:
            logger.warning(f"get() found no nodes, data.name='{type}-{name}' returning None")
            return None
        elif len(res) > 1:
            logger.warning(f"get() found many nodes (?), data.name='{type}-{name}' returning None")
            return None
        else:
            node = Node(res[0]['data']['type'], res[0]['data']['name'], dict(res[0]['data'].items()))
            logger.debug ("Query Successful:")
            logger.debug (node)

        node.set_data(node.properties.get('data', None))
        node.set_aux (node.properties.get('aux', None))

        return node

    @staticmethod
    def run(namespace, container_id, pipeline):
        """
        Runner for pipelined jobs
        """
        for func in pipeline:
            module = func[0]
            params = func[1]
            module (cohort_id=namespace, container_id=container_id, method_data=params)

    def runLocal(self, pipeline):
        """
        Run a pipeline in the main thread, blocking.

        :params: pipeline - an ordered list of (function, params) tuples to execute
        """
        self.run (self._namespace_id, self._name, pipeline)

    def runProcessPoolExecutor(self, pipeline, executor):
        """
        Use a process pool executor to run full pipelines in background

        :params: pipeline - an ordered list of (function, params) tuples to execute
        :params: executor - a ProcessPoolExecutor passed from a parent script
        """

        assert isinstance(executor, ProcessPoolExecutor)
        return executor.submit(self.run, self._namespace_id, self._name, pipeline)

    def runDaskDistributed(self, pipeline, client):
        """
        Submit functions to dask workers.
        Dask can track dependencies via a semaphore future, so we pass that explicitly and submit each function individually

        :params: pipeline - an ordered list of (function, params) tuples to execute
        :params: client - a dask client
        """
        from dask.distributed   import Client

        assert isinstance(client, Client)
        future = client.submit (bootstrap, self._name)
        for func in pipeline:
            module = func[0]
            params = func[1]
            future = client.submit (module, self._namespace_id, self._name, params, semaphore=future)
        return future


    def put(self, node: Node):
        """
        Adds a node to a temporary dictonary that will be used to save/commit nodes to the relevant databases
        If you add the same node under the same name, no change as the
        Decorates the node with the container's namespace

        :param: node - node object
        """
        assert isinstance(node, Node)
        assert self.isAttached()

        logger.info(f"Adding node: {node}")

        # Decorate with the container namespace
        node.set_namespace( self._namespace_id, self._name )
        node._datastore_id = self._datastore_id

        # Set node data object(s) only if there is a path and the object store is enabled
        node.objects = []
        if node.data is not None and self.params.get("OBJECT_STORE_ENABLED", False):
            node.properties['object_bucket'] = f"{self._bucket_id}"
            node.properties['object_folder'] = f"{self._name}/{node.name}"

            data_path = pathlib.Path( node.data )

            if data_path.is_file():
                node.objects.append( data_path )

            if data_path.is_dir():
                # TODO: enable extention in glob via something?
                for path in data_path.glob("*.*"):
                    node.objects.append(path)

            logger.info ("Node has %s pending object commits",  len(node.objects))

        # Set node aux object only if a path and the object store is enabled
        if node.aux is not None and self.params.get("OBJECT_STORE_ENABLED", False):
            node.properties['object_bucket'] = f"{self._bucket_id}"
            node.properties['object_folder'] = f"{self._name}/{node.name}"
            node.objects.append( pathlib.Path( node.aux ))
            logger.info ("Node has %s pending object commits",  len(node.objects))

        # Add to node commit dictonary
        logger.info ("Adding: %s", node.get_address())

        self._conn.query( f"""
            MATCH (container) WHERE id(container) = {node._datastore_id}
            MERGE (container)-[:HAS_DATA]->(da:{node.get_match_str()})
                ON MATCH  SET da = {node.get_map_str()}
                ON CREATE SET da = {node.get_map_str()}
            """ )

        if self.params.get("OBJECT_STORE_ENABLED", False):
            future_uploads = []
            executor = ThreadPoolExecutor(max_workers=4)

            object_bucket = node.properties.get("object_bucket")
            object_folder = node.properties.get("object_folder")
            for p in node.objects:
                future = executor.submit(self._client.fput_object, object_bucket, f"{object_folder}/{p.name}", p, part_size=250000000)
                future_uploads.append(future)

            n_count_futures = 0
            n_total_futures = len (future_uploads)
            for future in as_completed(future_uploads):
                try:
                    data = future.result()
                except:
                    logger.exception('Bad upload: generated an exception:')
                else:
                    n_count_futures += 1
                    if n_count_futures < 10: logger.info("Upload successful with etag: %s", data[0])
                    if n_count_futures < 1000 and n_count_futures % 100 == 0: logger.info("Uploaded [%s/%s]", n_count_futures, n_total_futures)
                    if n_count_futures % 1000 == 0: logger.info("Uploaded [%s/%s]", n_count_futures, n_total_futures)

            logger.info("Uploaded [%s/%s]", n_count_futures, n_total_futures)
            logger.info("Shutdown executor %s", executor)
            executor.shutdown()
        logger.info("Done saving all records!!")

    def add(self, *args):
        logger.warning ("Datastore.add() has been depreciated")
    def saveAll(self, *args):
        logger.warning ("Datastore.saveAll() has been depreciated")



# -*- coding: utf-8 -*-
"""
Created on October 17, 2019

@author: aukermaa@mskcc.org
"""
from luna.common.config import ConfigSet
import luna.common.constants as const


def test_table_name():
    c1 = ConfigSet(
        name=const.DATA_CFG,
        config_file="pyluna-common/tests/luna/common/testdata/data_ingestion_template_valid.yml",
    )

    assert const.TABLE_NAME(c1) == "CT_OV_16-158_CT_20201028"


def test_table_location():
    c1 = ConfigSet(
        name=const.DATA_CFG,
        config_file="pyluna-common/tests/luna/common/testdata/data_ingestion_template_valid.yml",
    )

    assert (
        const.TABLE_LOCATION(c1)
        == "pyluna-radiology/tests/luna/radiology/proxy_table/test_data/OV_16-158/tables/CT_OV_16-158_CT_20201028"
    )


def test_table_location_emptystring():
    c1 = ConfigSet(
        name=const.DATA_CFG,
        config_file="pyluna-common/tests/luna/common/testdata/data_ingestion_template_valid_empty_dataset.yml",
    )

    assert (
        const.TABLE_LOCATION(c1)
        == "pyluna-common/tests/luna/radiology/proxy_table/test_data/OV_16-158/tables/CT"
    )


def test_table_location_none():
    c1 = ConfigSet(
        name=const.DATA_CFG,
        config_file="pyluna-common/tests/luna/common/testdata/data_ingestion_template_valid_empty_dataset_2.yml",
    )

    assert (
        const.TABLE_LOCATION(c1)
        == "pyluna-common/tests/luna/radiology/proxy_table/test_data/OV_16-158/tables/CT"
    )


def test_project_location():
    c1 = ConfigSet(
        name=const.DATA_CFG,
        config_file="pyluna-common/tests/luna/common/testdata/data_ingestion_template_valid.yml",
    )

    assert (
        const.PROJECT_LOCATION(c1)
        == "pyluna-radiology/tests/luna/radiology/proxy_table/test_data/OV_16-158"
    )



'''
Created on November 02, 2020

@author: pashaa@mskcc.org
'''
import timeit


class CodeTimer:
    def __init__(self, logger, name=None):
        self.logger = logger
        self.name = " '"  + name + "'" if name else ''

    def __enter__(self):
        self.start = timeit.default_timer()

    def __exit__(self, exc_type, exc_value, traceback):
        self.took = (timeit.default_timer() - self.start)
        self.logger.info('Code block' + self.name + ' took: ' + str(self.took) + 's')
        if exc_type is not None:
            self.logger.exception("Exception raised during code execution:")



import scipy.stats
import numpy as np

def compute_stats_1d(vec, fx_name_prefix, n_percentiles=4):
    """ Computes 1d (histogram)-like summary statistics
    
    Args:
        vec (np.array): a 1-d vector input
        fx_name_prefix (str): Prefix for feature names
        n_percentiles (int): Number of percentiles to compute, default 4 = 0 (min), 25, 50, 75, 100 (max)
    
    Returns:
        dict: summary statistics
    """
    n, _, sm, sv, ss, sk = scipy.stats.describe(vec)
    # ln_params = scipy.stats.lognorm.fit(vec, floc=0)

    hist_features = {
        f'{fx_name_prefix}_nobs': n,
        f'{fx_name_prefix}_mean': sm,
        f'{fx_name_prefix}_variance': sv,
        f'{fx_name_prefix}_skewness': ss,
        f'{fx_name_prefix}_kurtosis': sk,
        # f'{fx_name_prefix}_lognorm_fit_p0': ln_params[0],
        # f'{fx_name_prefix}_lognorm_fit_p2': ln_params[2]
    }
    
    percentiles = np.linspace(0, 100, n_percentiles + 1)

    for percentile, value in zip(percentiles, np.percentile(vec, percentiles)):
        hist_features[f'{fx_name_prefix}_pct{int(percentile)}'] = value

    return hist_features


# -*- coding: utf-8 -*-
'''
Created on October 17, 2019

@author: aukermaa@mskcc.org
'''
import pytest

from luna.common.Node import Node, CONTAINER_TYPES, ALL_DATA_TYPES

def test_patient_create():
    create_string = Node("patient", "my_patient", properties={"namespace":"my_cohort", "Description":"a patient"}).get_create_str()
    assert "patient:globals" in create_string    
    assert "qualified_address: 'my_cohort::my_patient'"   in create_string
    assert "name"  in create_string
    assert "Description"   in create_string

def test_patient_match():
    match_string  = Node("patient", "my_patient", properties={"namespace":"my_cohort", "Description":"a patient"}).get_match_str()
    assert "patient" in match_string    
    assert "qualified_address: 'my_cohort::my_patient'"   in match_string
    assert "Description"  not in match_string


def test_cohort_create():
    create_string = Node("cohort", "my_cohort", properties={"Description":"a cohort"}).get_create_str()
    assert "cohort:globals" in create_string    
    assert "qualified_address: 'my_cohort::my_cohort'"  in create_string
    assert "name"  in create_string
    assert "Description"  in create_string

def test_cohort_match():
    match_string = Node("cohort", "my_cohort", properties={"Description":"a cohort"}).get_match_str()
    assert "cohort" in match_string    
    assert "qualified_address: 'my_cohort::my_cohort'"   in match_string
    assert "Description"  not in match_string


def test_metadata_create():
    properties = {}
    properties['namespace'] = "my_cohort"
    properties['dim'] = 3

    create_string = Node("VolumetricImage", "SCAN-001", properties=properties).get_create_str()
    assert "VolumetricImage:globals" in create_string    
    assert "qualified_address: 'my_cohort::VolumetricImage-SCAN-001'"  in create_string
    assert "dim"  in create_string

def test_metadata_match():
    properties = {}
    properties['namespace'] = "my_cohort"
    properties['dim'] = 3

    match_string = Node("VolumetricImage", "SCAN-002", properties=properties).get_match_str()
    assert "qualified_address: 'my_cohort::VolumetricImage-SCAN-002"  in match_string
    assert "dim"  not in match_string


def test_get_all_properties():
    properties = {}
    properties['namespace'] = "my_cohort"
    properties['dim'] = 3

    all_props = Node("VolumetricImage", "SCAN-001", properties=properties).get_all_props()
    assert "name" in all_props

def test_patient_no_namespace():
    node = Node("patient", "pid", properties={})
    assert node.properties['qualified_address'] == 'pid'

def test_cohort_wrong_properties():
    with pytest.raises(TypeError):
        Node("cohort", properties={"Description":"a cohort"})


def test_patient_bad_id():
    with pytest.raises(ValueError):
        Node("patient", "my:patient", properties={"namespace":"my_cohort", "Description":"a patient"})

def test_cohort_bad_id():
    with pytest.raises(ValueError):
        Node("cohort", "my:cohort", properties={"Description":"a cohort"})
      
@pytest.mark.parametrize(('node_type'),CONTAINER_TYPES)
def test_container_types(node_type):
    node = Node(node_type, "my_node")
    assert node.properties["type"] == node_type

@pytest.mark.parametrize(('node_type'),ALL_DATA_TYPES)
def test_container_types(node_type):
    node = Node(node_type, "my_node")
    assert node.properties["type"] == node_type
    assert node.properties["qualified_address"] == node_type + '-' + "my_node"
